(* $Id: godi_plan.ml 1182 2013-04-14 12:44:14Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_versions
open Godi_query
open Printf

exception Mission_impossible of string
exception Package_installation_not_unique of string

module StrMap = Map.Make(String);;

module StrSet = Set.Make(String);;

let max_steps = 100_000


let cat_weight =
  function
      `Installed_pkg _ -> 1
    | `Binary_pkg _ -> 2
    | `Source_pkg _ -> 3
    | `Wrapped_pkg _ -> 4
    | `Dispersed_pkg _ -> 5
;;


let create_pkg_hash ?(unique=false) pkgs =
  (* Creates a hash table mapping package names to lists of packages with
   * that name. The lists are sorted; the best package comes first.
   *)
  let ht = Hashtbl.create 100 in
  List.iter
    (fun pkg ->
       let name = (any_pkg pkg) # name in
       let old_list = try Hashtbl.find ht name with Not_found -> [] in
       if unique && old_list <> [] then
	 raise(Package_installation_not_unique name);
       Hashtbl.replace ht name (pkg :: old_list)
    )
    pkgs;
  (* Sort lists by version numbers, etc. *)
  Hashtbl.iter
    (fun name pkg_list ->
       let pkg_list' =
	 List.sort 
	   (fun pkg1 pkg2 ->
	      (* First sort by package type, then by version: *)
	      let w1 = cat_weight pkg1 in
	      let w2 = cat_weight pkg2 in
	      match w1 - w2 with
		  0 -> 
		    let v1 = (any_pkg pkg1)#version in
		    let v2 = (any_pkg pkg2)#version in
		    -(cmp_version_string v1 v2)
		| _ -> w1 - w2
	   )
	   pkg_list in
       Hashtbl.replace ht name pkg_list'
    )
    ht;
  ht
;;


let find_pkg ht name dep_list =
  (* Find the packages [name] that matches all conditions of [dep_list]: *)
  let candidates = 
    try Hashtbl.find ht name with Not_found -> [] in
  List.filter
    (fun pkg ->
       List.for_all
         (fun (name',cond) -> name'=name && matches_dep pkg (name, cond))
         dep_list)
    candidates
;;


let add_conflicts cflmap pkg =
  (* Updates the conflict map [cflmap] by adding the conflicts recorded
   * in the package description for [pkg], and returns the modified
   * conflict map
   *)
  let name = pkg_name pkg in
  let conflicts = pkg_conflicts pkg in
  List.fold_left
    (fun acc' (n,c) ->
       let l = try StrMap.find n acc' with Not_found -> [] in
       StrMap.add n ( (name, (n,c)) :: l ) acc'
    )
    cflmap
    conflicts
;;


let get_conflict_map pkgs =
  (* Returns a map of package names to conflict conditions, i.e.
   * the mapping
   *   name2 --> [...; (name1, (name2, cond)); ...]
   * means that the existence of the package name1 implies the conflict
   * condition that the version string of name2 must fulfill cond.
   *)
  List.fold_left add_conflicts StrMap.empty pkgs
;;


let conflicts lookup_pkg cflmap pkg =
  (* Returns the conflicts resulting by adding [pkg] to the set of selected
   * packages. It is checked whether there is a conflict for [pkg] in
   * [cflmap], and it is checked whether the conflict list of [pkg] 
   * includes a valid condition.
   *
   * [lookup_pkg] must be a function that looks up a package by name,
   * or raises [Not_found].
   *
   * The function returns a list of conflict conditions
   * [(name1, (name2, cond))] (see get_conflict_map).
   *)
  let name = pkg_name pkg in
  let cflconds = try StrMap.find name cflmap with Not_found -> [] in
  (* Are there other packages against [pkg]? *)
  List.filter
    (fun (carrier_pkg, cfldep) ->
       matches_dep pkg cfldep
    )
    cflconds @
  (* Is [pkg] against any other package? *)
  List.filter
    (fun (carrier_pkg, (cflname, cflcond as cfldep)) ->
       try
	 let pkg' = lookup_pkg cflname in
	 matches_dep pkg' cfldep
       with
	   Not_found -> false
    )
    ( List.map
	(fun cfldep ->  (name, cfldep))
	(pkg_conflicts pkg)
    )
;;

let print_conflict (carrier_pkg, cfldep) =
  "Package " ^ carrier_pkg ^ " lists conflict condition " ^ 
  print_dependency cfldep
;;

    
exception Dead_end;;


let rec try_in_order ?(onerror=fun () -> ()) f l =
  match l with
      x :: l' ->
	(try f x 
	 with Dead_end -> 
	   try_in_order f l')
    | [] ->
	onerror();
	raise Dead_end
;;


let rec accumulate_deps f set conflicts deps followup_action =
  match deps with
    | dep :: deps' ->
	let (_, _, (name,cond)) = dep in
	f 
	  set 
	  conflicts
	  name 
	  dep 
	  (fun set' -> accumulate_deps f set' conflicts deps' followup_action)
    | [] ->
	followup_action set
;;


type replacement_strategy =
    [ `Never
    | `Upgrade
    | `Any
    ]
;;


let print_violated_dependency (from_pkg_opt, _, dep) =
  match from_pkg_opt with
    | None ->
	"the requirement is " ^ print_dependency dep
    | Some pkg ->
	let t =
	  match pkgtype pkg with
	    | `Source -> "source"
	    | `Wrapped -> "wrapped source"
	    | `Binary -> "binary"
	    | `Installed -> "installed"
	    | `Dispersed -> "dispersed" in

	t ^ 
	" package " ^ pkg_fullname_at pkg ^ " requires " ^ print_dependency dep
;;


let rec count_non_source_pkgs l =
  match l with
    | (None, _, _)  :: l' ->
	count_non_source_pkgs l'
    | (Some pkg, _, _)  :: l' ->
	if pkgtype pkg <> `Source then
	  1 + count_non_source_pkgs l'
	else
	  count_non_source_pkgs l'
    | [] ->
	0
;;


let debug = false (* true *)

let analyze_requirements 
      ?(replace = `Never) ?(select_by_dep = [])
      ~available ~selected () =
  let avail_ht = create_pkg_hash available in
  let sel_ht = create_pkg_hash ~unique:true selected in

  (* add_package: This is a backtracking algorithm trying to find the
   * best package match. It adds the best package [name] that has been
   * selected because of the dependency [dep] to the already selected
   * [set], and, if this is successful, calls [followup_action] with the
   * new set. The [followup_action] may raise Dead_end to signal that
   * this did not lead to a good end. In this case, the second-best package
   * [name] is selected instead, and so on, until a match is found.
   *
   * The sets are StrMap.t values (i.e. maps in reality) mapping names
   * to the pairs (pkg,deps), where pkg is the package and deps is the
   * list of collected dependencies that must all hold.
   *
   * The packages of [selected] are preferred over the packages of 
   * [available].
   *
   * I hope nobody gets really into the complexity of this algorithm...
   * I am sure there is a better one.
   *)

  let steps = ref 0 in

  let emap = ref StrMap.empty in

  let error m =
    if StrMap.mem m !emap then
      let n = StrMap.find m !emap in
      emap := StrMap.add m (n+1) !emap
    else
      emap := StrMap.add m 1 !emap in

  let debug_add set name xdep =
    let (_,_,dep) = xdep in
    eprintf "TEST %s DEP %s FOR %s\n"
      name
      (print_dependency dep)
      (String.concat "," 
         (List.map
            (fun (_,(pkg,_)) -> pkg_fullname_at pkg) 
            (StrMap.bindings set)
         )
      ) in

  let rec add_package set conflmap name xdep
                      (followup_action : 'a -> 'a) =
    if debug then debug_add set name xdep;
    if !steps > max_steps then raise Dead_end;
    incr steps;
    let member = StrMap.mem name set in
    let (from_pkg_opt, from_selpkg_name, dep) = xdep in
    if member then (
      (* Case: the package is already selected. We only check whether
       * this package also fulfills the new condition dep. If not,
       * raise Dead_end. If yes, add the new conditions to the set, and
       * continue with the followup_action.
       *
       * We need not to check for conflicts here, as the package is
       * already incorporated into the set of selected packages.
       *)
      let (pkg, stored_deps) = StrMap.find name set in
      if matches_dep pkg dep then (
        if debug then eprintf "  -> match\n";        
	let set' = StrMap.add name (pkg, xdep :: stored_deps) set in
	followup_action set'
      )
      else (
        if debug then eprintf "  -> contra\n";        
        error
	  (sprintf
	     "Version conflict in the dependency chain for package %s:\n  %s,\n  but %s cannot be installed in this version"
	     from_selpkg_name
	     (print_violated_dependency xdep)
	     name);
	raise Dead_end
      )
    )
    else (
      if debug then eprintf "   -> search\n";        
      (* Case: the package is unknown. Get the list of possible packages
       * that fulfill dep, and try them in turn, until a path is found.
       * Every package has its own dependencies which are tried as
       * a followup_action.
       *)
      let pkg_list1 =
	find_pkg sel_ht name [dep] in
      let pkg_list2 = 
	find_pkg avail_ht name [dep] in
      let pkg_list = 
	match replace with
	    `Never ->
	      (* First try [selected], then [available] *)
	      if pkg_list1 = [] then
		find_pkg avail_ht name [dep]
	      else
		pkg_list1
	  | `Upgrade ->
	      (* First try [selected], then upgrades in [available] *)
	      if pkg_list1 = [] then
		pkg_list2
	      else
		let p = List.hd pkg_list1 in
		pkg_list1 @ 
		(List.filter (fun p' -> not(newer_package p p')) pkg_list2)
                  (* NB. "not(newer_package p p')" includes the case that
                     the existing version is rebuilt: An ancestor of p
                     might already be upgraded, and as a consequence we need
                     the option to rebuild existing packages.
                   *)
	  | `Any ->
	      (* Try both *)
	      pkg_list1 @ pkg_list2
      in
      (* The best package is the first in pkg_list (because the packages
       * have been sorted by create_pkg_hash).
       *)
      if pkg_list = [] then (
        if debug then eprintf "  -> search empty\n";
        let msg =
	  (sprintf "Cannot figure out how to select package: %s%s.%s" 
	     (print_dependency dep)
	     (if from_selpkg_name = fst dep then "" else
		"\nThis package is needed as a dependency of " ^ 
                  from_selpkg_name
	     )
             (if replace=`Never then
                "\nIt is also possible that enabling upgrades would solve this."
              else
                ""
             )
          ) in
        if debug then eprintf "ERROR: %s\n" msg;
        error msg
      );
      if debug then eprintf "  -> search list: %s\n"
        (String.concat "," (List.map pkg_fullname_at pkg_list));
      try_in_order
	(fun pkg ->
	   (* Try pkg *)
           if debug then eprintf " -> next %s\n" (pkg_fullname_at pkg);
	   let set' =
	     StrMap.add name (pkg, [xdep]) set in
	   (* Check whether there are conflicts. *)
	   let conflmap' = add_conflicts conflmap pkg in
	   let lookup_pkg n =
	     try fst(StrMap.find n set)
	     with Not_found ->
	       match find_pkg sel_ht n [] with
		   []  -> raise Not_found
		 | [p] -> p
		 | _   -> assert false
	   in
	   let cfllist = conflicts lookup_pkg conflmap' pkg in
	   if cfllist <> [] then (
             if debug then eprintf " -> conflict\n";
	     error (sprintf
		      "Conflicts:\n  %s"
		      (String.concat "\n  "
			 (List.map print_conflict cfllist)));
	     raise Dead_end
	   );

	   (* Add the dependencies in turn, finally do followup_action: *)
	   let deps' = (any_pkg pkg)#dependencies @
		       ( match pkg with
			     `Source_pkg _ -> 
			       (any_pkg pkg)#build_dependencies
			   | _  -> [] 
		       ) in
	   let xdeps' = 
	     List.map (fun dep -> ((Some pkg), from_selpkg_name, dep)) deps' in
	   accumulate_deps add_package set' conflmap' xdeps' followup_action
	)
	pkg_list
    )
  in

  (* Create initial map of conflict conditions: *)
  let conflicts = get_conflict_map selected in

  (* selected: first try the packages that need to be rebuilt anyway.
     This reduces the search space.
   *)
  let selected1, selected_other =
    List.partition
      (fun pkg ->
         match pkg_user_status pkg with
           | `Build | `Build_dep _ -> true
           | _ -> false
      )
      selected in

  (* For added packages do it the other way round: *)
  let selected2, selected3 =
    List.partition
      (fun pkg ->
         match pkg_user_status pkg with
           | `Add | `Add_dep _ -> true
           | _ -> false
      )
      selected_other in

  (* Add the packages of [selected] in turn: *)
  let set =
    try
      accumulate_deps 
	add_package
	StrMap.empty
	conflicts
	(List.map 
	   (fun pkg -> 
	      let name = (any_pkg pkg)#name in
	      (None, name, parse_dependency (  name ^ "-*" ) )
	   )
	   (selected1 @ selected2 @ selected3)
         @
           (* We assume here this is only for adding packages, so it is
              ok to do it last
            *)
           (List.map
              (fun (name,cond) -> (None,name,(name,cond)))
              select_by_dep
           )
	)
	(fun x -> x) 
    with
	Dead_end ->
	  (* It is not possible to fulfill all conditions! *)
          let errors = StrMap.bindings !emap in
          let errors' = 
            List.sort (fun (_,a) (_,b) -> a-b) errors in
          let msg =
            String.concat "\n\n" (List.map fst errors') in
          raise(Mission_impossible
                  ( "While searching for a solution, the planner ran into \
                     the following problems:\n\n" ^ msg))
  in
  (* Make a list from the set: *)
  StrMap.fold
    (fun name (pkg,_) l -> pkg::l)
    set
    []
;;


let get_dependent_packages ?(include_build_deps=false) ~active ~selected () =
  let actv_ht = create_pkg_hash active in
  let succ_ht = Hashtbl.create 100 in
      (* maps name of package to list of successors *)
  let visited_ht = Hashtbl.create 100 in
      (* a package is "visited" if all successors are known *)

  let init_ht() =
    List.iter
      (fun pkg ->
	 let p = any_pkg pkg in
	 let name = p # name in
	 let deps =
	   if include_build_deps then
	     p # dependencies @ p # build_dependencies
	   else
	     p # dependencies in
	 List.iter
	   (fun (dep_name,_) ->
	      let l = 
		try Hashtbl.find succ_ht dep_name 
		with Not_found -> StrSet.empty in
	      if not (StrSet.mem name l) then (
		Hashtbl.replace succ_ht dep_name (StrSet.add name l)
	      )
	   )
	   deps
      )
      active
  in

  let rec traverse pkg =
    let p = any_pkg pkg in
    let name = p # name in

    if not (Hashtbl.mem visited_ht pkg) then (

      let succ_set =
	try 
	  Hashtbl.find succ_ht name
	with 
	    Not_found -> StrSet.empty in
      
      let succ_names =
	StrSet.elements succ_set in
      
      let succ_pkgs =
	List.flatten
	  (List.map
	     (fun n ->
		let dep = Godi_versions.any_dependency n in
		find_pkg actv_ht n [ dep ])
	     succ_names
	  ) in

      Hashtbl.replace visited_ht pkg ();

      List.iter traverse succ_pkgs;

      let all_succs =
	List.fold_left
	  (fun acc n ->
	     let s = 
	       try Hashtbl.find succ_ht n with Not_found -> StrSet.empty  in
	     StrSet.union acc s
	  )
	  succ_set
	  succ_names in
      Hashtbl.replace succ_ht name all_succs;
    )
  in

  (* Initialize succ_ht with the direct dependencies: *)
  init_ht();
  (* Find the closure of succ_ht: *)
  List.iter traverse active;
  (* Return the result, which is contained in succ_ht: *)
  let all_succs =
    List.fold_left
      (fun acc pkg ->
	 let p = any_pkg pkg in
	 let succs =
	   try Hashtbl.find succ_ht p#name with Not_found -> StrSet.empty in
	 StrSet.union acc succs)
      StrSet.empty
      selected in
  List.flatten
    (List.map
       (fun name ->
	  let dep = Godi_versions.any_dependency name in
	  find_pkg actv_ht name [ dep ])
       (StrSet.elements all_succs)
    )
;;


let get_required_packages ~active ~selected () =
  let actv_ht = create_pkg_hash active in
  let visited_ht = Hashtbl.create 100 in
      (* a package is "visited" if all predecessors are known *)

  let rec closure pkg =
    (* Computes the set of prerequisite packages for [pkg]. Returns that as
     * StrSet of package names.
     *)
    let p = any_pkg pkg in
    (* let name = p # name in *)

    if not (Hashtbl.mem visited_ht pkg) then (

      let pred_names =
	List.map
	  (fun (name,_) -> name )
	  p # dependencies in

      let pred_pkgs =
	List.flatten
	  (List.map
	     (fun n ->
		let dep = Godi_versions.any_dependency n in
		find_pkg actv_ht n [ dep ])
	     pred_names
	  ) in

      Hashtbl.replace visited_ht pkg ();

      let set_list =
	List.map (fun p -> StrSet.add (pkg_name p) (closure p)) pred_pkgs in

      let all_preds =
	List.fold_left 
	  StrSet.union
	  StrSet.empty
	  set_list in
      all_preds
    )
    else
      StrSet.empty
  in

  let set_list = List.map closure selected in
  let names =
    StrSet.elements
      (List.fold_left 
	 StrSet.union
	 StrSet.empty
	 set_list) in
  List.flatten
    (List.map
       (fun n ->
	  let dep = Godi_versions.any_dependency n in
	  find_pkg actv_ht n [ dep ])
       names
    )
;;


let sort_packages pkgs =
  let pkgs_ht = create_pkg_hash pkgs in
  let visited = Hashtbl.create 100 in

  let rec antecedents pkg =
    if not(Hashtbl.mem visited pkg) then (
      let deps = (any_pkg pkg)#dependencies @
		 ( match pkg with
		       `Source_pkg _ -> 
			 (any_pkg pkg)#build_dependencies
		     | _  -> [] 
		 ) in
      Hashtbl.replace visited pkg ();
      List.fold_left
	(fun acc (name,cond) ->
	   let dep_pkgs = find_pkg pkgs_ht name [name,cond] in
	   let dep_pkgs' =
	     List.fold_left
	       (fun acc' dep_pkg ->
		  let dep_pkgs' = antecedents dep_pkg in
		  acc' @ dep_pkgs'
	       )
	       []
	       dep_pkgs in
	   acc @ dep_pkgs'
	)
	[]
	deps
      @ [ pkg ]
    )
    else []
  in

  List.fold_left
    (fun acc pkg ->
       let pkg' = antecedents pkg in
       acc @ pkg'
    )
    []
    pkgs
;;


let make_set l =
  PkgMap.make (List.map (fun x -> x, ()) l)
;;


let update_plan ?(replace = `Never) ?(select_by_dep = [])
                ~iteration pkgs =
  let count = ref 0 in
  let iteration_ht = Hashtbl.create 100 in

  let get_build_iteration pkg =
    try
      match Hashtbl.find iteration_ht pkg with
	  `B i -> i
	| _    -> raise Not_found
      with
	  Not_found -> iteration 
  in

  let get_remove_iteration pkg =
    try
      match Hashtbl.find iteration_ht pkg with
	  `R i -> i
	| _    -> raise Not_found
      with
	  Not_found -> iteration 
  in

  let get_add_iteration pkg =
    try
      match Hashtbl.find iteration_ht pkg with
	  `A i -> i
	| _    -> raise Not_found
      with
	  Not_found -> iteration 
  in

  (* First set all `Remove_dep, `Build_dep, and `Add_dep packages to `Keep.
   * We remember the old iteration counter (iteration_ht).
   *)
  List.iter
    (fun pkg ->
       let p = any_pkg pkg in
       match p#user_status with
	   `Remove_dep i ->
	     Hashtbl.add iteration_ht pkg (`R i);
	     p#set_user_status `Keep
	 | `Build_dep i ->
	     Hashtbl.add iteration_ht pkg (`B i);
	     p#set_user_status `Keep
	 | `Add_dep i ->
	     Hashtbl.add iteration_ht pkg (`A i);
	     p#set_user_status `Keep
	 | _ ->
	     ()
    )
    pkgs;

  (* Partition pkgs: *)

  let inst_pkgs, pkgs' =
    List.partition
      (function `Installed_pkg _ -> true | _ -> false)
      pkgs in

  let src_pkgs, bin_pkgs =
    List.partition
      (function `Source_pkg _ -> true | _ -> false)
      pkgs' in

  let src_set = make_set src_pkgs in
  let bin_set = make_set bin_pkgs in

  (* Compute the closure of all packages to remove: *)

  (* Get installed packages that need to be removed by deps: *)

  let rm_pkgs =
    List.filter to_be_removed inst_pkgs in

  let rm_inst_pkgs =
    get_dependent_packages ~active:inst_pkgs ~selected:rm_pkgs () in

  let rm_set =
    PkgMap.union (make_set rm_pkgs) (make_set rm_inst_pkgs) in

  List.iter
    (fun pkg ->
       let p = any_pkg pkg in
       if p # user_status = `Keep then (
	 let i = get_remove_iteration pkg in
	 p#set_user_status (`Remove_dep i);
	 if i = iteration then
	   incr count
       )
    )
    rm_inst_pkgs;

  (* After the removal, the set of installed packages is smaller: *)

  let inst_pkgs' = 
    List.filter (fun pkg -> not (PkgMap.mem pkg rm_set)) inst_pkgs in
  let inst_set' =
    make_set inst_pkgs' in

  (* Get installed packages that need to be rebuilt: 
   *
   * (~active is inst_pkgs, not inst_pkgs', because the question is which
   * packages were previously installed.)
   *)

  let inst_pkgs_to_keep =
    (* Installed packages the user wants to keep *)
    List.filter
      (fun p -> pkg_user_status p <> `Remove)
      inst_pkgs in

  let build_pkgs =
    List.filter to_be_built src_pkgs in

  let sel_inst_pkgs =
    get_dependent_packages ~active:inst_pkgs_to_keep ~selected:build_pkgs () in

  (* Find the corresponding source packages, and select them: *)

  List.iter
    (fun pkg ->
       let name = (any_pkg pkg) # name in
       try
	 let src_pkg =
	   match PkgMap.find_by ~name src_set with
	     | []       -> raise Not_found
	     | [ p, _ ] -> p
	     | _        -> assert false in
	 let s = source_pkg src_pkg in
	 ( match s#user_status with
	     | `Keep ->
		 let i = get_build_iteration src_pkg in
		 s # set_user_status (`Build_dep i);
		 if i = iteration then
		   incr count
	     | _ ->
		 ()
	 )
       with
	   Not_found -> ()
    )
    sel_inst_pkgs;

  (* Do the requirements analysis: *)

  (* Get the list of packages available for installation: *)

  let available =
    match replace with
      | `Never ->
          (* Non-installed source packages, including packages that were
           * removed by the removal pass above:
           *)
          List.filter
            (fun pkg ->
	       let name = (any_pkg pkg) # name in 
	       not (PkgMap.mem_by ~name inst_set')
            )
            src_pkgs
          @
            (* Non-installed binary packages, ... *)
            List.filter
            (fun pkg ->
	       let name = (any_pkg pkg) # name in 
	       not (PkgMap.mem_by ~name inst_set')
            )
            bin_pkgs
      | _ ->
          (* If we allow upgrades/downgrades, we cannot exclude the
             packages anymore than are already installed
           *)
          src_pkgs @ bin_pkgs 
  in

  (* Determine the list of selected packages:
   * - Installed pkgs that are not removed, excluding rebuilt sources
   * - Source pkgs selected for build
   * - Binary pkgs selected for addition
   *
   * Packages must be unique, i.e. it is illegal to have two packages
   * with the same name.
   *)

  let selected = 
    List.filter
      (fun ip ->
	 let name = (any_pkg ip)#name in
	 let src_list = List.map fst (PkgMap.find_by ~name src_set) in
	 let bin_list = List.map fst (PkgMap.find_by ~name bin_set) in
	 not(List.exists to_be_built src_list) &&
	 not(List.exists to_be_added bin_list)
      )
      inst_pkgs'
    @
    List.filter to_be_built src_pkgs
    @
    List.filter to_be_added bin_pkgs
  in

  (* Because packages to rebuild may occur both in inst_pkgs' and in
   * build_pkgs, the question is which object should take precedence.
   * Of course, the source package, as the dependencies may have changed.
   *)

   (* The case that a newly selected package requires a new version
    * of an already selected package is controlled by the [replace]
    * argument.
    *)

  let reqs =
    analyze_requirements
      ~select_by_dep
      ~replace
      ~available
      ~selected
      () 
  in

  (* Select the missing packages: *)
  List.iter
    (fun req ->
       match req with
	   `Source_pkg p ->
	     ( match p#user_status with
		 | `Keep ->
		     let i = get_build_iteration req in
		     p # set_user_status (`Build_dep i);
		     if i = iteration then
		       incr count
		 | _ ->
		     ()
	     )
	 | `Binary_pkg p ->
	     ( match p#user_status with
		 | `Keep ->
		     let i = get_add_iteration req in
		     p # set_user_status (`Add_dep i);
		     if i = iteration then
		       incr count
		 | _ ->
		     ()
	     )
	 | _ ->
	     ()
    )
    reqs;

  (* Finally, schedule installed packages for removal, if the corresponding
   * source package will be built, or the corresponding binary package
   * will be added:
   *)

  List.iter
    (fun pkg ->
       let p = any_pkg pkg in
       let name = p#name in
       let src_list = List.map fst (PkgMap.find_by ~name src_set) in
       let bin_list = List.map fst (PkgMap.find_by ~name bin_set) in
       if List.exists to_be_built src_list then (
	 let i = get_remove_iteration pkg in
	 p#set_user_status (`Remove_dep i);
	 if i = iteration then
	   incr count
       )
       else
	 if List.exists to_be_added bin_list then (
	   let i = get_remove_iteration pkg in
	   p#set_user_status (`Remove_dep i);
	   if i = iteration then
	     incr count
	 )
    )
    inst_pkgs';

  !count
;;
